#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include "mpi.h"
#include "stdarg.h"
#include <sys/time.h>
#include <unistd.h>


#ifdef offsetof
#undef offsetof
#endif
#define offsetof(s_type,field)	((size_t) &(((s_type *)(void *)0)->field))

static int dodebug = 0;

static
int
prnt (int myrank, const char *fmt, ...) 
{
	va_list	ap;
	char    str[1024];
	char    *s;

	if (!dodebug) return 1;
	va_start (ap, fmt);
	sprintf (str, "[%d]: ", myrank);
	s = index (str, ':') + 2;
	vsnprintf (s, 1000, fmt, ap);
	s[1001] = 0;
	fputs (str, stderr);
	fflush (stderr);
	va_end (ap);
	return 1;
}


static double t1;

int
micro_start ()
{
	t1 = MPI_Wtime ();
}


long
micro_get ()
{
    return (MPI_Wtime () - t1) * 1000 * 1000;
}


extern int MPID_SMI_emulate;


int
main (argc, argv)
	int		argc;
	char	** argv;
{
	int				myrank=-1, numprocs;
	int				i, j, k;
	MPI_Win			win;
	char			* base = NULL;
	char			* buffer = NULL;
	int				size = 100, numruns = 20, endebug = 0;
	int				testnum = 1;
	MPI_Datatype	dtype, dtype_help;
	int				numofdata;
	double			secs[2][10000];
	int				numgaps = 0;
	int				*blocklens;
	MPI_Aint		*disps;
	MPI_Datatype	*types;
	MPI_Aint		dtype_extent;
	int				dtype_size;
	int				testtype=1;
	double			schnittx, schnitty, olddiff, diff, divisor,
					y11, y12, y21, y22;
	int				hasold;
	int				maxsize = 2000, minsize = 20, grain = 5;
	int				emul;
	float			fragment[10000];

	prnt (myrank, "MPI_Init\n");
	MPI_Init (&argc, &argv);

	for (i=1; i<argc; i++) {
		if (!strcmp (argv[i], "-n")) {
			if (++i < argc) 
				numruns = atoi (argv[i]);
		} else if (!strcmp (argv[i], "-s")) {
			if (++i < argc)
				maxsize = atoi (argv[i]);
		} else if (!strcmp (argv[i], "-m")) {
			if (++i < argc)
				minsize = atoi (argv[i]);
		} else if (!strcmp (argv[i], "-g")) {
			if (++i < argc)
				grain = atoi (argv[i]);
		} else if (!strcmp (argv[i], "-q")) {
			dodebug = 0;
			endebug = 0;
		} else if (!strcmp (argv[i], "-d")) {
			dodebug = 1;
			endebug = 1;
		} else if (!strcmp (argv[i], "-t")) {
			testtype = 0;
		}
	}

	prnt (myrank, "MPI_Comm_rank\n");
	MPI_Comm_rank (MPI_COMM_WORLD, &myrank);
	prnt (myrank, "MPI_Comm_size\n");
	MPI_Comm_size (MPI_COMM_WORLD, &numprocs);

	prnt (myrank, "MPI_Alloc_mem\n");
	MPI_Alloc_mem (maxsize * 4 * sizeof (int), MPI_INFO_NULL, &base);
	if (!base) {
		fprintf (stderr, "No shared memory\n");
		return -1; 
	}
	buffer = malloc (maxsize * 4 * sizeof (int));
	if (!buffer) {
		fprintf (stderr, "Out of memory\n");
		return -2;
	}
	disps = malloc (maxsize * sizeof (MPI_Aint));
	blocklens = malloc (maxsize * sizeof (int));
	types = malloc (maxsize * sizeof (MPI_Datatype));

	prnt (myrank, "MPI_Win_create\n");
	MPI_Win_create (base, size, 0, MPI_INFO_NULL, MPI_COMM_WORLD, &win);

	prnt (myrank, "put_same_type test no. 1\n");

	MPI_Win_fence (0, win);
	if (myrank == 1) {
		for (size=20; size <= maxsize; size+=5) {
			for (emul = 0; emul <= 1; emul++) {
			if (emul == 0) 
				MPID_SMI_emulate = 0;
			else
				MPID_SMI_emulate = 1;
			for (numgaps = 1; numgaps <= size; numgaps++) {
				for (i=0; i<size; i++) {
					blocklens[i] = 1;
					if (testtype) {
						if (i <= numgaps) 
							disps[i] = 2 * sizeof (int) * i;
						else
							disps[i] = (2 * numgaps + (i-numgaps)) * 
													sizeof (int);
						types[i] = MPI_INT;
					} else {
						disps[i] = sizeof (int) * i;
						if (i < numgaps) 
							types[i] = MPI_CHAR;
						else
							types[i] = MPI_INT;
					}
				}

				MPI_Type_struct (size, blocklens, disps, types, &dtype);
				MPI_Type_commit (&dtype);
				MPI_Type_extent (dtype, &dtype_extent);
				MPI_Type_size	(dtype, &dtype_size);
		
				fragment[numgaps] = 100. * ((float)dtype_size / 
											(float)dtype_extent);

				micro_start ();
				for (j=0; j<numruns; j++) {
					MPI_Put (buffer, 1, dtype, 0, 0, 1, dtype, win);
				}
				secs[emul][numgaps] = micro_get ();

			}}
			hasold = 0;
			olddiff = 0.;
			for (i=1; i<=size; i++) {
				for (j=1; j<=size; j++) {
					diff = secs[0][i] - secs[1][i];
					if (diff == 0. || hasold && (
								diff < 0. && olddiff > 0. ||
								diff > 0. && olddiff < 0.)) {
#if 0
						y11 = secs[0][i-1];
						y12 = secs[0][i];
						y21 = secs[1][i-1];
						y22 = secs[1][i];
						divisor = y12 - y11 - y22 + y21;
						schnittx = (y21 - y11 + ((double)(i-1))*divisor) / 
										divisor;
						schnitty = (y12 - y11) * schnittx + y11 - 
										((double)(i-1)) * (y12 - y11);
#endif
						printf ("%d %2.2f\n", size * 4, fragment[i]);
						fflush (stdin);
						goto endloop;
					}
					olddiff = diff;
					hasold = 1;
				}
			}
endloop:
		}
	}

	
	prnt (myrank, "MPI_Win_fence\n");
	MPI_Win_fence (0, win);
	prnt (myrank, "MPI_Win_free\n");
	MPI_Win_free (&win);
	prnt (myrank, "MPI_Finalize\n");
	MPI_Finalize ();
	prnt (myrank, "Good bye!\n");

	return 0;
}


























/*
 * Overrides for XEmacs and vim so that we get a uniform tabbing style.
 * XEmacs/vim will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-basic-offset: 4
 * tab-width: 4
 * End:
 * vim:tw=0:ts=4:wm=0:
 */
