#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include "mpi.h"
#include "stdarg.h"
#include <sys/time.h>
#include <unistd.h>


#ifdef offsetof
#undef offsetof
#endif
#define offsetof(s_type,field)	((size_t) &(((s_type *)(void *)0)->field))

static int dodebug = 0;

static
int
prnt (int myrank, const char *fmt, ...) 
{
	va_list	ap;
	char    str[1024];
	char    *s;

	if (!dodebug) return 1;
	va_start (ap, fmt);
	sprintf (str, "[%d]: ", myrank);
	s = index (str, ':') + 2;
	vsnprintf (s, 1000, fmt, ap);
	s[1001] = 0;
	fputs (str, stderr);
	fflush (stderr);
	va_end (ap);
	return 1;
}


static double t1;

int
micro_start ()
{
	t1 = MPI_Wtime ();
}


long
micro_get ()
{
    return (MPI_Wtime () - t1) * 1000 * 1000;
}





int
main (argc, argv)
	int		argc;
	char	** argv;
{
	int				myrank=-1, numprocs;
	int				i, j, k;
	MPI_Win			win;
	char			* base = NULL;
	char			* buffer = NULL;
	int				size = 100, numruns = 1, endebug = 0;
	int				testnum = 1;
	MPI_Datatype	dtype, dtype_help;
	int				numofdata;
	double			secs;
	int				numgaps = 0;
	int				*blocklens;
	MPI_Aint		*disps;
	MPI_Datatype	*types;
	MPI_Aint		dtype_extent;
	int				dtype_size;
	int				testtype=0;

	prnt (myrank, "MPI_Init\n");
	MPI_Init (&argc, &argv);

	for (i=1; i<argc; i++) {
		if (!strcmp (argv[i], "-n")) {
			if (++i < argc) 
				numruns = atoi (argv[i]);
		} else if (!strcmp (argv[i], "-s")) {
			if (++i < argc)
				size = atoi (argv[i]);
		} else if (!strcmp (argv[i], "-q")) {
			dodebug = 0;
			endebug = 0;
		} else if (!strcmp (argv[i], "-d")) {
			dodebug = 1;
			endebug = 1;
		} else if (!strcmp (argv[i], "-t")) {
			testtype = 1;
		}
	}

	prnt (myrank, "MPI_Comm_rank\n");
	MPI_Comm_rank (MPI_COMM_WORLD, &myrank);
	prnt (myrank, "MPI_Comm_size\n");
	MPI_Comm_size (MPI_COMM_WORLD, &numprocs);

	prnt (myrank, "MPI_Alloc_mem\n");
	MPI_Alloc_mem (size * 4 * sizeof (int), MPI_INFO_NULL, &base);
	if (!base) {
		fprintf (stderr, "No shared memory\n");
		return -1; 
	}
	buffer = malloc (size * 4 * sizeof (int));
	if (!buffer) {
		fprintf (stderr, "Out of memory\n");
		return -2;
	}
	disps = malloc (size * sizeof (MPI_Aint));
	blocklens = malloc (size * sizeof (int));
	types = malloc (size * sizeof (MPI_Datatype));

	prnt (myrank, "MPI_Win_create\n");
	MPI_Win_create (base, size, 0, MPI_INFO_NULL, MPI_COMM_WORLD, &win);

	prnt (myrank, "put_same_type test no. 1\n");

	MPI_Win_fence (0, win);
	if (myrank == 1) {
		for (numgaps = 1; numgaps <= size; numgaps++) {
			prnt (myrank, "create dtype\n");
			for (i=0; i<size; i++) {
				blocklens[i] = 1;
				if (testtype) {
					if (i <= numgaps) 
						disps[i] = 2 * sizeof (int) * i;
					else
						disps[i] = (2 * numgaps + (i-numgaps)) * sizeof (int);
					types[i] = MPI_INT;
				} else {
					disps[i] = sizeof (int) * i;
					if (i < numgaps) 
						types[i] = MPI_CHAR;
					else
						types[i] = MPI_INT;
				}
			}

			MPI_Type_struct (size, blocklens, disps, types, &dtype);
			MPI_Type_commit (&dtype);
			MPI_Type_extent (dtype, &dtype_extent);
			MPI_Type_size	(dtype, &dtype_size);
		


			micro_start ();
			for (j=0; j<numruns; j++) {
				MPI_Put (buffer, 1, dtype, 0, 0, 1, dtype, win);
			}
			secs = micro_get ();

			printf ("%2.2f %8.3lf\n", 100. * ((float)dtype_size / 
										(float)dtype_extent),
										 secs / (double) numruns);
		}
	}

	
	prnt (myrank, "MPI_Win_fence\n");
	MPI_Win_fence (0, win);
	prnt (myrank, "MPI_Win_free\n");
	MPI_Win_free (&win);
	prnt (myrank, "MPI_Finalize\n");
	MPI_Finalize ();
	prnt (myrank, "Good bye!\n");

	return 0;
}


























/*
 * Overrides for XEmacs and vim so that we get a uniform tabbing style.
 * XEmacs/vim will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-basic-offset: 4
 * tab-width: 4
 * End:
 * vim:tw=0:ts=4:wm=0:
 */
