#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include "mpi.h"
#include "stdarg.h"

#ifdef offsetof
#undef offsetof
#endif
#define offsetof(s_type,field)	((size_t) &(((s_type *)(void *)0)->field))

static int dodebug = 1;

static
int
prnt (int myrank, const char *fmt, ...) 
{
	va_list	ap;
	char    str[1024];
	char    *s;

	if (!dodebug) return 1;
	va_start (ap, fmt);
	sprintf (str, "[%d]: ", myrank);
	s = index (str, ':') + 2;
	vsnprintf (s, 1000, fmt, ap);
	s[1001] = 0;
	fputs (str, stderr);
	fflush (stderr);
	va_end (ap);
	return 1;
}



int
main (argc, argv)
	int		argc;
	char	** argv;
{
	int				myrank=-1, numprocs, mysize;
	int				i, j, k;
	MPI_Win			win;
	char			* base = NULL;
	char			buffer [2048];
	int				size = 1024, numruns = 1, endebug = 1;
	int				testnum = 1;
	MPI_Datatype	dtype, dtype_help;
	struct test2_t {
		int		a;
		char	b;
		int		c[20];
	} 				* test2_val;
	struct test3_t1 {
		char	a;
		int		b;
		char	c;
	};
	struct test3_t {
		struct test3_t1	a[10];
		char			b;
	} 				* test3_val;
	int				blocklens[3];
	MPI_Aint		disps[3];
	MPI_Datatype	types[3];
	MPI_Aint		dtype_extent;
	int				dtype_size;

	prnt (myrank, "MPI_Init\n");
	MPI_Init (&argc, &argv);

	for (i=1; i<argc; i++) {
		if (!strcmp (argv[i], "-n")) {
			if (++i < argc) 
				numruns = atoi (argv[i]);
		} else if (!strcmp (argv[i], "-s")) {
			if (++i < argc)
				size = atoi (argv[i]);
		} else if (!strcmp (argv[i], "-q")) {
			dodebug = 0;
			endebug = 0;
		} else if (!strcmp (argv[i], "-t")) {
			if (++i < argc)
				testnum = atoi (argv[i]);
		}
	}

	prnt (myrank, "MPI_Comm_rank\n");
	MPI_Comm_rank (MPI_COMM_WORLD, &myrank);
	prnt (myrank, "MPI_Comm_size\n");
	MPI_Comm_size (MPI_COMM_WORLD, &numprocs);

	if (3 * sizeof (struct test2_t) * numprocs > size) {
		size = 3 * sizeof (struct test2_t) * numprocs;
		prnt (myrank, "warning: size extended to %d\n", size);
	}
	mysize = size + (size/64) * (myrank+1);
	prnt (myrank, "MPI_Alloc_mem\n");
	MPI_Alloc_mem (mysize, MPI_INFO_NULL, &base);
	if (!base) {
		fprintf (stderr, "No shared memory\n");
		return -1; 
	}

	prnt (myrank, "MPI_Win_create\n");
	MPI_Win_create (base, mysize, sizeof (struct test2_t), 
					MPI_INFO_NULL, MPI_COMM_WORLD, &win);

	prnt (myrank, "put_same_type test no. 1\n");

	prnt (myrank, "create dtype\n");
	blocklens[0] = 1;
	blocklens[1] = 1;
	blocklens[2] = 20;
	disps[0] = offsetof (struct test2_t, a);
	disps[1] = offsetof (struct test2_t, b);
	disps[2] = offsetof (struct test2_t, c);
	types[0] = MPI_INT;
	types[1] = MPI_CHAR;
	types[2] = MPI_INT;
	MPI_Type_struct (3, blocklens, disps, types, &dtype);
	MPI_Type_commit (&dtype);
	MPI_Type_extent (dtype, &dtype_extent);
	MPI_Type_size	(dtype, &dtype_size);

	prnt (myrank, "size (test2_t) = %d\nextent (test2_t) = %d\n",
					dtype_size, dtype_extent);
	prnt (myrank, "size/extent = %2.2f\n", (float)dtype_size / 
											(float)dtype_extent);
	for (j=0; j<numruns; j++) {
		MPI_Win_fence (0, win);
		test2_val = (struct test2_t *) buffer;
		for (i=0; i<3; i++) {
			test2_val[i].a = 113;
			test2_val[i].b = 'a';
			test2_val[i].c[0] = i;
			test2_val[i].c[1] = j;
			test2_val[i].c[2] = myrank;
			test2_val[i].c[3] = 113;
		}
		MPI_Win_fence (0, win);
		MPI_Win_lock (MPI_LOCK_EXCLUSIVE, 0, 0, win);
		MPI_Put (buffer, 3, dtype, 0, 3*myrank, 3, dtype, win);
		MPI_Win_unlock (0, win);
		MPI_Win_fence (0, win);

		if (myrank == 0) {
			for (i=0; i<numprocs; i++) {
				test2_val = (struct test2_t *) base + 3*i;
				printf ("proc %d has written: ", i);
				for (k=0; k<3; k++) 
					printf ("(%d%c:%d/%d/%d:%d) ", 
							test2_val[k].a, test2_val[k].b,
							test2_val[k].c[0], test2_val[k].c[1],
							test2_val[k].c[2], test2_val[k].c[3]);
				printf ("\n");
				fflush (stdout);
			}
		}
	}
	
	prnt (myrank, "MPI_Win_fence\n");
	MPI_Win_fence (0, win);
	prnt (myrank, "MPI_Win_free\n");
	MPI_Win_free (&win);
	prnt (myrank, "MPI_Finalize\n");
	MPI_Finalize ();
	prnt (myrank, "Good bye!\n");

	return 0;
}


























/*
 * Overrides for XEmacs and vim so that we get a uniform tabbing style.
 * XEmacs/vim will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-basic-offset: 4
 * tab-width: 4
 * End:
 * vim:tw=0:ts=4:wm=0:
 */
